#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

#include "memory.h"
#include "fail.h"

static void* memory_start = NULL;
static void* memory_end = NULL;
static void* free_boundary = NULL;

char* memory_identity = "no free";

void memory_setup(size_t total_size) {
  memory_start = malloc(total_size);
  assert(memory_start != NULL);
  memory_end = memory_start + total_size;
}

void memory_cleanup() {
  assert(memory_start != NULL);
  free(memory_start);
  memory_start = memory_end = free_boundary = NULL;
}

void* memory_get_start() {
  return memory_start;
}

void* memory_get_end() {
  return memory_end;
}

void memory_set_heap_start(void* heap_start) {
  assert(free_boundary == NULL);
  free_boundary = heap_start;
}

void* memory_allocate(int words) {
  assert(free_boundary != NULL);

  size_t size = words * sizeof(intptr_t);
  if (size == 0)
    size = sizeof(intptr_t);
  void* res = free_boundary;
  free_boundary += size;
  if (free_boundary > memory_end)
    fail("no memory left (%zd bytes requested)", size);
  return res;
}
